from web3cat.view.wireframes.abstract import *
from web3cat.view.wireframes.erc20 import *
from web3cat.view.wireframes.chainlink import *
from web3cat.view.wireframes.eth import *
from web3cat.view.wireframes.portfolio import *
